% Hierarchical construction of the Schur complement of a FEM matrix.
% A quadtree is used, and boxes are merged by pairs or two.
% Dense linear algebra is used throughout, so this is O(N^1.5).

function main21()

rng('default')
rng(0)

nlevel = 3;
nleaf  = 5;
nside  = nleaf*(2^nlevel);
N      = nside*nside;
acc    = 1e-10;
fprintf(1,'\n=== Parameters: nlevel=%2d   nleaf=%3d   nside=%4d   ntot=%d\n',...
        nlevel,nleaf,nside,N)

% Construct the stiffness matrix (this is a sparse matrix).
flag_matrix = 'rcon';
flag_anchor = 1;
A = LOCAL_get_A(nside,nside,flag_matrix,flag_anchor);

fprintf(1,'\n=== Construct the Schur complements\n')
tic
S = LOCAL_get_Schur_dense_hierarchical(A,nlevel,nleaf);
fprintf(1,'    Time required = %0.2f (sec)\n',toc)
fprintf(1,'    Memory required for full Schur complement = %d (KB)\n',...
               round(LOCAL_get_bytes(S)/1024))

fprintf(1,'\n=== Compressing the Schur complement into HSS form\n')
NODES_comp = LOCAL_compress_HSS_fsym_brute(S,acc,50);
%NODES_comp = compress_HSS_nsym_brute(S,acc,50);
fprintf(1,'    Memory required = %d (KB)   mem =  %0.1f (words/node)\n',...
        round(LOCAL_get_bytes(NODES_comp)/1024),LOCAL_get_bytes(NODES_comp)/(8*4*nside))
LOCAL_rank_statistics_from_HSS(NODES_comp);

%%% Verify the solution.
fprintf(1,'\n=== Verifying the solution\n')
[Jrim,Jint] = LOCAL_get_Jrim_Jint_for_square(nside);
f       = zeros(length(Jint) + length(Jrim),1);
f(Jrim) = randn(length(Jrim),1);
u_ref   = A\f;
u       = S\f(Jrim);
fprintf(1,'    Absolute error from a test solve = %9.2e\n',norm(u - u_ref(Jrim)))
fprintf(1,'    Relative error from a test solve = %9.2e\n',norm(u - u_ref(Jrim))/norm(u_ref(Jrim)))
fprintf(1,'\n')

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function A = LOCAL_get_A(nside1,nside2,flag_matrix,flag_anchor)

if (nargin < 4)
  flag_anchor = 0;
end

if (nargin == 1)
  nside2 = nside1;
end

BARS  = LOCAL_get_links(nside1,nside2);
nbars = size(BARS,2);

if strcmp(flag_matrix,'orig')
    
  BARS(3,:) = ones(1,nbars);
  
elseif strcmp(flag_matrix,'rcon')
    
  BARS(3,:) = 1 + rand(1,nbars);
  
elseif strcmp(flag_matrix,'rcut')
    
  p = 0.04;
  BARS(3,:) = (rand(1,nbars) > p);
  
elseif strcmp(flag_matrix,'peri')
  
  [XX1,XX2] = meshgrid(0:(nside1-1),0:(nside2-1));
  xx        = [reshape(XX1,1,nside1*nside2);reshape(XX2,1,nside1*nside2)];
  h         = 1/(max([nside1,nside2])-1);
  Ncell     = round(min(nside1,nside2)/25);
  XX1sc     = h*xx(1,BARS(1,:));
  XX2sc     = h*xx(2,BARS(1,:));
  YY1sc     = h*xx(1,BARS(2,:));
  YY2sc     = h*xx(2,BARS(2,:));
  ZZ1sc     = 0.5*(XX1sc+YY1sc);
  ZZ2sc     = 0.5*(XX2sc+YY2sc);
  BARS(3,:) = 1 - 0.9*((cos(Ncell*pi*ZZ1sc).*cos(Ncell*pi*ZZ2sc)).^2);
    
elseif strcmp(flag_matrix,'crac')

  % We create a list of bars to be removed:
  n1a = round(0.4*nside1);
  n1b = round(0.6*nside1);
  n2a = round(0.4*nside2);
  n2b = round(0.7*nside2);
  gap = 4;
  
  ii1 = n1a*nside2 + (1:n2a);
  jj1 = ii1 + nside2;
  ii2 = n1a*nside2 + n2a + 1 + nside2*(1:(n1b-n1a));
  jj2 = ii2 - 1;
  ii3 = n1b*nside2 + 1 + (n2a:n2b);
  jj3 = ii3 + nside2;
  ii4 = (n1b-gap)*nside2 + (nside2:(-1):(n2a+gap+2));
  jj4 = ii4-nside2;
  ii5 = (n1a-1)*nside2 + (n2a+gap+1) + nside2*(1:(n1b-n1a-gap));
  jj5 = ii5+1;
  
  ii = [ii1,ii2,ii3,ii4,ii5];
  jj = [jj1,jj2,jj3,jj4,jj5];
  
  ni   = length(ii);
  BARS(3,:) = ones(1,nbars);
  BARS = [BARS,[ii;jj;-ones(1,ni)]];
      
end

if (flag_anchor == 0)
    
  A = sparse([BARS(1,:), BARS(1,:), BARS(2,:), BARS(2,:)],...
             [BARS(1,:), BARS(2,:), BARS(1,:), BARS(2,:)],...
             [BARS(3,:),-BARS(3,:),-BARS(3,:), BARS(3,:)]);
         
else
    
  ii_rim = [1:nside1,...
            (nside1+1):nside1:(nside1*(nside2-2)+1),...
            (2*nside1):nside1:(nside1*(nside2-1)),...
             nside1*(nside2-1) + (1:nside1)];
     
  A = sparse([BARS(1,:), BARS(1,:), BARS(2,:), BARS(2,:),ii_rim],...
             [BARS(1,:), BARS(2,:), BARS(1,:), BARS(2,:),ii_rim],...
             [BARS(3,:),-BARS(3,:),-BARS(3,:), BARS(3,:),ones(1,2*(nside1+nside2-2))]);
    
end

if strcmp(flag_matrix,'crac')
  [ii,jj,aa] = find(A);
  ind = find(abs(aa) < 1e-12);
  aa(ind) = zeros(size(ind));
  A = sparse(ii,jj,aa);
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This supplementary functions returns a list of all links in a
% rectangular grid of size n1 x n2.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function BARS = LOCAL_get_links(n1,n2)

if (nargin == 1)
  n2 = n1;
end

BARS  = zeros(3,2*n1*n2 - n1 - n2);

ndone = 0;
for i1 = 1:(n1-1)
  for i2 = 1:(n2-1)
    j_c = n2*(i1-1) + i2;
    j_n = n2*(i1-1) + i2 + 1;
    j_e = n2*i1 + i2;
    BARS(1:2,ndone + (1:2)) = [j_c, j_c; j_n, j_e];
    ndone = ndone + 2;
  end
  j_c = n2*i1;
  j_e = n2*(i1+1);
  BARS(1:2,ndone + 1) = [j_c; j_e];
  ndone = ndone + 1;
end
for i2 = 1:(n2-1)
  j_c = n2*(n1-1) + i2;
  j_n = n2*(n1-1) + i2 + 1;
  BARS(1:2,ndone + 1) = [j_c; j_n];
  ndone = ndone + 1;
end

BARS(3,:) = ones(1,ndone);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function S = LOCAL_get_Schur_dense_hierarchical(A,nlevel,nleaf)

% Construct the data structure.
NODES = LOCAL_get_NODES(nlevel,nleaf);

% Construct the list specifying the order in which the boxes are to be
% scanned. While following this list, any box encountered will either be 
% a leaf box, or a box whose children have already been processed.
box_list = LOCAL_scan_tree(NODES);

% Create the data structure holding the compressed Schur complements.
% Note that at any given time, most slots will be empty as any Schur
% complement that is no longer needed will be erased.
S_comp_STORE = cell(1,size(NODES,2));

for icount = 1:length(box_list)
    
  ibox = box_list(icount);

  if (NODES{2,ibox} == nlevel) % ibox is a leaf
    
    S_comp_STORE{ibox} = LOCAL_get_S_comp_leaf(A,NODES,ibox);

  else % ibox is not a leaf
    
    %%% Merge sons 1 and 3:
    ison1 = NODES{ 4,ibox};
    ison3 = NODES{ 6,ibox};
    SL = S_comp_STORE{ison1};
    SR = S_comp_STORE{ison3};
    S_comp_STORE{ison1} = [];
    S_comp_STORE{ison3} = [];
    SD = LOCAL_merge_two_horizontal(SL,SR,A);
      
    %%% Merge sons 2 and 4:
    ison2 = NODES{ 5,ibox};
    ison4 = NODES{ 7,ibox};
    SL = S_comp_STORE{ison2};
    SR = S_comp_STORE{ison4};
    S_comp_STORE{ison2} = [];
    S_comp_STORE{ison4} = [];
    SU = LOCAL_merge_two_horizontal(SL,SR,A);

    % Merge the DOWN and the UP Schur complements.
    S_comp_STORE{ibox} = LOCAL_merge_two_vertical(SD,SU,A);

  end
  
end

% Extract the Schur complement for the top node in "normal" matrix format.
S = LOCAL_get_S_from_S_comp(S_comp_STORE{1});

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function mem = LOCAL_get_bytes(X)

tmp = whos('X');

mem = tmp.bytes;

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [kvec,nvec] = LOCAL_rank_statistics_from_HSS(NODES,flag_print)

if (nargin == 1)
  flag_print = 1;
end

nboxes  = size(NODES,2);
nlevels = 0;
for ibox = 1:nboxes
  nlevels = max(nlevels,NODES{2,ibox});
end

k_list = cell(1,nlevels);
n_list = cell(1,nlevels);

for ibox = 2:size(NODES,2)
  ilevel    = NODES{2,ibox};
  k         = NODES{9,ibox};
  n         = NODES{7,ibox};
  k_list{ilevel} = [k_list{ilevel},k];
  n_list{ilevel} = [n_list{ilevel},n];
end

kvec = zeros(1,nlevels);
nvec = zeros(1,nlevels);

for ilevel = 2:nlevels
  k_mean = mean(k_list{ilevel});
  n_mean = mean(n_list{ilevel});
  kvec(ilevel) = k_mean;
  nvec(ilevel) = n_mean;
  if (flag_print == 1)
    fprintf(1,'    level = %2d   n_mean = %6.1f   k_mean = %6.1f\n',...
            ilevel,n_mean,k_mean)
  end
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [Jrim,Jint] = LOCAL_get_Jrim_Jint_for_square(nside)

Jrim = [1 + nside*(0:(nside-1)),...
        1 + nside*(nside-1) + (1:(nside-1)),...
        nside*(nside-1):(-nside):(2*nside),...
        nside:(-1):2];
    
Jint = zeros(1,(nside-2)*(nside-2));

for i = 1:(nside-2)
  Jint((nside-2)*(i-1) + (1:(nside-2))) = 1 + i*nside + (1:(nside-2));
end
      
return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function NODES = LOCAL_compress_HSS_fsym_brute(A,acc,nbox_max)

ntot = size(A,1);

% Compute the tree strcture.
NODES  = LOCAL_get_tree(ntot,nbox_max);
nboxes = size(NODES,2);

% Compress all nodes, going from smaller to larger.
for ibox = nboxes:(-1):2
  ind_offd = [1:(NODES{6,ibox}-1),(NODES{6,ibox}+NODES{7,ibox}):ntot];
  if ( (NODES{4,ibox}<=0) && (NODES{5,ibox}<=0) ) % ibox has no sons.
    ind_skel = NODES{6,ibox} - 1 + (1:NODES{7,ibox});
  elseif ( (NODES{4,ibox} > 0) && (NODES{5,ibox}<= 0) ) % ibox has a left son
    ison1    = NODES{4,ibox};
    ind_skel = NODES{20,ison1};
  elseif ( (NODES{4,ibox}<=0) && (NODES{5,ibox}>0) ) % ibox has a right son
    ison2    = NODES{5,ibox};
    ind_skel = NODES{20,ison2};
  else % ibox has two sons
    ison1    = NODES{4,ibox};
    ison2    = NODES{5,ibox};
    ind_skel = [NODES{20,ison1},NODES{20,ison2}];
  end
  A21 = A(ind_offd,ind_skel);
  % Compute the skeletons.
  [T,J] = LOCAL_id_decomp(A21,acc);
  k = size(T,1);
  % Record the outgoing skeletons:
  NODES{20,ibox} = ind_skel(J(1:k));
  NODES{30,ibox} = T;
  NODES{31,ibox} = J;
  % Record nskel and kskel
  NODES{ 8,ibox} = length(ind_skel);
  NODES{ 9,ibox} = k;
end

% Construct the matrices representing self-interactions on the leaves.
for ibox = nboxes:(-1):2
  if ( (NODES{4,ibox}<=0) && (NODES{5,ibox}<=0) )
    ind = NODES{6,ibox} - 1 + (1:NODES{7,ibox});
    NODES{40,ibox} = A(ind,ind);
  end
end

% Construct the matrices for sibling interactions.
for ibox = nboxes:(-1):1
  if ( (NODES{4,ibox}>0) && (NODES{5,ibox}>0) )
    ison1 = NODES{4,ibox};
    ison2 = NODES{5,ibox};
    NODES{46,ison1} = A(NODES{20,ison1},NODES{20,ison2});
  end
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function NODES = LOCAL_get_tree(ntot,nbox_max)

NODES_0 = cell(7,10000);

% Construct the top node.
NODES_0{2,1} = 0;
NODES_0{3,1} = NaN;
NODES_0{4,1} = -1;
NODES_0{5,1} = -1;
NODES_0{6,1} =  1;
NODES_0{7,1} = ntot;

% Create smaller nodes via hierarchical subdivision.
% We sweep one level at a time.
ibox_last = 0;
ncreated  = 1;
ilevel    = 0;
while (ncreated > 0)
  ibox_first = ibox_last + 1;
  ibox_last  = ibox_last + ncreated;
  ncreated   = 0;
  for ibox = ibox_first:ibox_last
    nbox = NODES_0{7,ibox};
    if (nbox > nbox_max)
      if (mod(nbox,2) == 2)
        nleft  = round(nbox/2);
      elseif (randn(1) > 0)
        nleft = ceil(nbox/2);
      else
        nleft = floor(nbox/2);
      end
      nright = nbox - nleft;
      ison1  = ibox_last + ncreated + 1;
      ison2  = ibox_last + ncreated + 2;
      NODES_0{4,ibox}      = ison1;
      NODES_0{5,ibox}      = ison2;
      NODES_0{2,ison1} = ilevel+1;
      NODES_0{2,ison2} = ilevel+1;
      NODES_0{3,ison1} = ibox;
      NODES_0{3,ison2} = ibox;
      NODES_0{4,ison1} = -1;
      NODES_0{4,ison2} = -1;
      NODES_0{5,ison1} = -1;
      NODES_0{5,ison2} = -1;
      NODES_0{6,ison1} = NODES_0{6,ibox};
      NODES_0{6,ison2} = NODES_0{6,ibox} + nleft;
      NODES_0{7,ison1} = nleft;
      NODES_0{7,ison2} = nright;
      ncreated = ncreated + 2;
    end
  end
  ilevel = ilevel + 1;
end

% This whole section is devoted to the simple task of
% decreasing the size of NODES_0.
% The operation NODES_0 = NODES_0{:,1:nboxes} apparently
% does not work for "cell" structures.
nboxes = ibox_last;
NODES = cell(50,nboxes);
for ibox = 1:nboxes
  for j = 1:size(NODES_0,1)
    NODES{j,ibox} = NODES_0{j,ibox};
  end
end
clear NODES_0

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function NODES = LOCAL_get_NODES(nlevel,nleaf)

nside  = nleaf*(2^nlevel);
ncells = (4^(nlevel+1) - 1)/3;
NODES  = cell(20,ncells);
NODES{ 2,1} = 0;
NODES{ 3,1} = NaN;
NODES{ 4,1} = 2;
NODES{ 5,1} = 3;
NODES{ 6,1} = 4;
NODES{ 7,1} = 5;
NODES{ 9,1} = nside;
NODES{10,1} = 1;
NODES{11,1} = 1;
ncreated    = 1;
for ilevel = 0:(nlevel-1)
  for ibox = (4^ilevel-1)/3 + (1:(4^ilevel))
    ison1 = ncreated + 1;
    ison2 = ncreated + 2;
    ison3 = ncreated + 3;
    ison4 = ncreated + 4;
    NODES{ 4,ibox } = ison1;
    NODES{ 5,ibox } = ison2;
    NODES{ 6,ibox } = ison3;
    NODES{ 7,ibox } = ison4;
    NODES{ 2,ison1} = ilevel+1;
    NODES{ 2,ison2} = ilevel+1;
    NODES{ 2,ison3} = ilevel+1;
    NODES{ 2,ison4} = ilevel+1;
    NODES{ 3,ison1} = ibox;
    NODES{ 3,ison2} = ibox;
    NODES{ 3,ison3} = ibox;
    NODES{ 3,ison4} = ibox;
    NODES{ 4,ison1} = -1;
    NODES{ 4,ison2} = -1;
    NODES{ 4,ison3} = -1;
    NODES{ 4,ison4} = -1;
    NODES{ 5,ison1} = -1;
    NODES{ 5,ison2} = -1;
    NODES{ 5,ison3} = -1;
    NODES{ 5,ison4} = -1;
    NODES{ 6,ison1} = -1;
    NODES{ 6,ison2} = -1;
    NODES{ 6,ison3} = -1;
    NODES{ 6,ison4} = -1;
    NODES{ 7,ison1} = -1;
    NODES{ 7,ison2} = -1;
    NODES{ 7,ison3} = -1;
    NODES{ 8,ison4} = -1;
    NODES{ 9,ison1} = nleaf*(2^(nlevel-ilevel-1));
    NODES{ 9,ison2} = nleaf*(2^(nlevel-ilevel-1));
    NODES{ 9,ison3} = nleaf*(2^(nlevel-ilevel-1));
    NODES{ 9,ison4} = nleaf*(2^(nlevel-ilevel-1));
    NODES{10,ison1} = 2*NODES{10,ibox}-1;
    NODES{11,ison1} = 2*NODES{11,ibox}-1;
    NODES{10,ison2} = 2*NODES{10,ibox}-1;
    NODES{11,ison2} = 2*NODES{11,ibox}  ;
    NODES{10,ison3} = 2*NODES{10,ibox}  ;
    NODES{11,ison3} = 2*NODES{11,ibox}-1;
    NODES{10,ison4} = 2*NODES{10,ibox}  ;
    NODES{11,ison4} = 2*NODES{11,ibox}  ;
    ncreated = ncreated + 4;
  end
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function box_list = LOCAL_scan_tree(NODES)
box_list          = zeros(1,size(NODES,2));
nfound            = 0;
[box_list,nfound] = LOCAL_scan_tree_rec(NODES,box_list,nfound,1);
return
        
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [box_list,nfound] = LOCAL_scan_tree_rec(NODES,box_list,nfound,ibox)
ison1 = NODES{4,ibox};
ison2 = NODES{5,ibox};
ison3 = NODES{6,ibox};
ison4 = NODES{7,ibox};
if (ison1 > -1)
  [box_list,nfound] = LOCAL_scan_tree_rec(NODES,box_list,nfound,ison1);
end
if (ison2 > -1)
  [box_list,nfound] = LOCAL_scan_tree_rec(NODES,box_list,nfound,ison2);
end
if (ison3 > -1)
  [box_list,nfound] = LOCAL_scan_tree_rec(NODES,box_list,nfound,ison3);
end
if (ison4 > -1)
  [box_list,nfound] = LOCAL_scan_tree_rec(NODES,box_list,nfound,ison4);
end
box_list(nfound+1) = ibox;
nfound             = nfound+1;
return
        
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function S_comp = LOCAL_get_S_comp_leaf(A,NODES,ibox)
    
[Jsw,Js,Jse,Je,Jne,Jn,Jnw,Jw] = LOCAL_get_J_pieces(NODES,ibox);
Jint                          = LOCAL_get_Jint(NODES,ibox);
Jc                            = [Jsw,Jse,Jne,Jnw];

T_int_s = A(Jint,Jint)\A(Jint,Js);
T_int_e = A(Jint,Jint)\A(Jint,Je);
T_int_n = A(Jint,Jint)\A(Jint,Jn);
T_int_w = A(Jint,Jint)\A(Jint,Jw);

S_comp      = cell(6,5);
S_comp{1,1} = full(A(Js,Js) - A(Js,Jint)*T_int_s);
S_comp{1,2} = full(         - A(Js,Jint)*T_int_e);
S_comp{1,3} = full(         - A(Js,Jint)*T_int_n);
S_comp{1,4} = full(         - A(Js,Jint)*T_int_w);
S_comp{1,5} = full(A(Js,Jc));

S_comp{2,1} = full(         - A(Je,Jint)*T_int_s);
S_comp{2,2} = full(A(Je,Je) - A(Je,Jint)*T_int_e);
S_comp{2,3} = full(         - A(Je,Jint)*T_int_n);
S_comp{2,4} = full(         - A(Je,Jint)*T_int_w);
S_comp{2,5} = full(A(Je,Jc));

S_comp{3,1} = full(         - A(Jn,Jint)*T_int_s);
S_comp{3,2} = full(         - A(Jn,Jint)*T_int_e);
S_comp{3,3} = full(A(Jn,Jn) - A(Jn,Jint)*T_int_n);
S_comp{3,4} = full(         - A(Jn,Jint)*T_int_w);
S_comp{3,5} = full(A(Jn,Jc));

S_comp{4,1} = full(         - A(Jw,Jint)*T_int_s);
S_comp{4,2} = full(         - A(Jw,Jint)*T_int_e);
S_comp{4,3} = full(         - A(Jw,Jint)*T_int_n);
S_comp{4,4} = full(A(Jw,Jw) - A(Jw,Jint)*T_int_w);
S_comp{4,5} = full(A(Jw,Jc));

S_comp{5,1} = full(A(Jc,Js) - A(Jc,Jint)*T_int_s);
S_comp{5,2} = full(A(Jc,Je) - A(Jc,Jint)*T_int_e);
S_comp{5,3} = full(A(Jc,Jn) - A(Jc,Jint)*T_int_n);
S_comp{5,4} = full(A(Jc,Jw) - A(Jc,Jint)*T_int_w);
S_comp{5,5} = full(A(Jc,Jc));

S_comp{6,1} = Js;
S_comp{6,2} = Je;
S_comp{6,3} = Jn;
S_comp{6,4} = Jw;
S_comp{6,5} = [Jsw,Jse,Jne,Jnw];

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function Snew_comp = LOCAL_merge_two_horizontal(SL,SR,A)

% Extract certain index vectors.
% (These are needed to retrieve/construct needed elements of A.)
JLe = SL{6,2};
JLc = SL{6,5};
JRw = SR{6,4};
JRc = SR{6,5};

% Construct the relevant components of the CENTER matrix.
X_aa     = SL{2,2};
X_bb     = SR{4,4};
X_aa_inv = inv(X_aa);
X_bb_inv = inv(X_bb);
Y_aa     = inv(X_aa - full(A(JLe,JRw)*(X_bb_inv*A(JRw,JLe))));
Y_bb     = inv(X_bb - full(A(JRw,JLe)*(X_aa_inv*A(JLe,JRw))));
Y_ab_rel = -full(X_aa_inv*A(JLe,JRw));
Y_ba_rel = -full(X_bb_inv*A(JRw,JLe));

%%% Compute influence from LEFT-SOUTH
F_a        = Y_aa*SL{2,1};
F_b        = Y_ba_rel*F_a;
Snew_LS_LS = SL{1,1} - SL{1,2}*F_a;
Snew_LN_LS = SL{3,1} - SL{3,2}*F_a;
Snew_LW_LS = SL{4,1} - SL{4,2}*F_a;
Snew_LC_LS = SL{5,1} - SL{5,2}*F_a;
Snew_RS_LS =         - SR{1,4}*F_b;
Snew_RE_LS =         - SR{2,4}*F_b;
Snew_RN_LS =         - SR{3,4}*F_b;
Snew_RC_LS =         - SR{5,4}*F_b;

%%% Compute influence from LEFT-NORTH
F_a        = Y_aa*SL{2,3};
F_b        = Y_ba_rel*F_a;
Snew_LS_LN = SL{1,3} - SL{1,2}*F_a;
Snew_LN_LN = SL{3,3} - SL{3,2}*F_a;
Snew_LW_LN = SL{4,3} - SL{4,2}*F_a;
Snew_LC_LN = SL{5,3} - SL{5,2}*F_a;
Snew_RS_LN =         - SR{1,4}*F_b;
Snew_RE_LN =         - SR{2,4}*F_b;
Snew_RN_LN =         - SR{3,4}*F_b;
Snew_RC_LN =         - SR{5,4}*F_b;

%%% Compute influence from LEFT-WEST
F_a        = Y_aa*SL{2,4};
F_b        = Y_ba_rel*F_a;
Snew_LS_LW = SL{1,4} - SL{1,2}*F_a;
Snew_LN_LW = SL{3,4} - SL{3,2}*F_a;
Snew_LW_LW = SL{4,4} - SL{4,2}*F_a;
Snew_LC_LW = SL{5,4} - SL{5,2}*F_a;
Snew_RS_LW =         - SR{1,4}*F_b;
Snew_RE_LW =         - SR{2,4}*F_b;
Snew_RN_LW =         - SR{3,4}*F_b;
Snew_RC_LW =         - SR{5,4}*F_b;

%%% Compute influence from LEFT-CENTER
F_a        = Y_aa*SL{2,5};
F_b        = Y_ba_rel*F_a;
Snew_LS_LC = SL{1,5}      - SL{1,2}*F_a;
Snew_LN_LC = SL{3,5}      - SL{3,2}*F_a;
Snew_LW_LC = SL{4,5}      - SL{4,2}*F_a;
Snew_LC_LC = SL{5,5}      - SL{5,2}*F_a;
Snew_RS_LC =              - SR{1,4}*F_b;
Snew_RE_LC =              - SR{2,4}*F_b;
Snew_RN_LC =              - SR{3,4}*F_b;
Snew_RC_LC = A(JRc,JLc) - SR{5,4}*F_b;

%%% Compute influence from RIGHT-SOUTH
F_b        = Y_bb*SR{4,1};
F_a        = Y_ab_rel*F_b;
Snew_LS_RS =         - SL{1,2}*F_a;
Snew_LN_RS =         - SL{3,2}*F_a;
Snew_LW_RS =         - SL{4,2}*F_a;
Snew_LC_RS =         - SL{5,2}*F_a;
Snew_RS_RS = SR{1,1} - SR{1,4}*F_b;
Snew_RE_RS = SR{2,1} - SR{2,4}*F_b;
Snew_RN_RS = SR{3,1} - SR{3,4}*F_b;
Snew_RC_RS = SR{5,1} - SR{5,4}*F_b;

%%% Compute influence from RIGHT-EAST
F_b        = Y_bb*SR{4,2};
F_a        = Y_ab_rel*F_b;
Snew_LS_RE =         - SL{1,2}*F_a;
Snew_LN_RE =         - SL{3,2}*F_a;
Snew_LW_RE =         - SL{4,2}*F_a;
Snew_LC_RE =         - SL{5,2}*F_a;
Snew_RS_RE = SR{1,2} - SR{1,4}*F_b;
Snew_RE_RE = SR{2,2} - SR{2,4}*F_b;
Snew_RN_RE = SR{3,2} - SR{3,4}*F_b;
Snew_RC_RE = SR{5,2} - SR{5,4}*F_b;

%%% Compute influence from RIGHT-NORTH
F_b        = Y_bb*SR{4,3};
F_a        = Y_ab_rel*F_b;
Snew_LS_RN =         - SL{1,2}*F_a;
Snew_LN_RN =         - SL{3,2}*F_a;
Snew_LW_RN =         - SL{4,2}*F_a;
Snew_LC_RN =         - SL{5,2}*F_a;
Snew_RS_RN = SR{1,3} - SR{1,4}*F_b;
Snew_RE_RN = SR{2,3} - SR{2,4}*F_b;
Snew_RN_RN = SR{3,3} - SR{3,4}*F_b;
Snew_RC_RN = SR{5,3} - SR{5,4}*F_b;

%%% Compute influence from RIGHT-CENTER
F_b        = Y_bb*SR{4,5};
F_a        = Y_ab_rel*F_b;
Snew_LS_RC =         -    SL{1,2}*F_a;
Snew_LN_RC =            - SL{3,2}*F_a;
Snew_LW_RC =            - SL{4,2}*F_a;
Snew_LC_RC = A(JLc,JRc) - SL{5,2}*F_a;
Snew_RS_RC = SR{1,5}    - SR{1,4}*F_b;
Snew_RE_RC = SR{2,5}    - SR{2,4}*F_b;
Snew_RN_RC = SR{3,5}    - SR{3,4}*F_b;
Snew_RC_RC = SR{5,5}    - SR{5,4}*F_b;

%%% Assemble the Schur structure for the whole matrix.
Snew_comp = cell(6,5);

Snew_comp{1,1} = [Snew_LS_LS,     Snew_LS_LC(:,2),Snew_LS_RC(:,1),Snew_LS_RS      ;...
                  Snew_LC_LS(2,:),Snew_LC_LC(2,2),Snew_LC_RC(2,1),Snew_LC_RS(2,:) ;...
                  Snew_RC_LS(1,:),Snew_RC_LC(1,2),Snew_RC_RC(1,1),Snew_RC_RS(1,:) ;...
                  Snew_RS_LS,     Snew_RS_LC(:,2),Snew_RS_RC(:,1),Snew_RS_RS     ];
Snew_comp{1,2} = [Snew_LS_RE     ;...
                  Snew_LC_RE(2,:);... 
                  Snew_RC_RE(1,:);... 
                  Snew_RS_RE];
Snew_comp{1,3} = [Snew_LS_RN,     Snew_LS_RC(:,4),Snew_LS_LC(:,3),Snew_LS_LN      ;...
                  Snew_LC_RN(2,:),Snew_LC_RC(2,4),Snew_LC_LC(2,3),Snew_LC_LN(2,:) ;...
                  Snew_RC_RN(1,:),Snew_RC_RC(1,4),Snew_RC_LC(1,3),Snew_RC_LN(1,:) ;...
                  Snew_RS_RN,     Snew_RS_RC(:,4),Snew_RS_LC(:,3),Snew_RS_LN     ];
Snew_comp{1,4} = [Snew_LS_LW     ;...
                  Snew_LC_LW(2,:);... 
                  Snew_RC_LW(1,:);... 
                  Snew_RS_LW];
Snew_comp{1,5} = [Snew_LS_LC(:,1),Snew_LS_RC(:,2:3),Snew_LS_LC(:,4);...
                  Snew_LC_LC(2,1),Snew_LC_RC(2,2:3),Snew_LC_LC(2,4);...
                  Snew_RC_LC(1,1),Snew_RC_RC(1,2:3),Snew_RC_LC(1,4);...
                  Snew_RS_LC(:,1),Snew_RS_RC(:,2:3),Snew_RS_LC(:,4)];

Snew_comp{2,1} = [Snew_RE_LS,Snew_RE_LC(:,2),Snew_RE_RC(:,1),Snew_RE_RS];
Snew_comp{2,2} = Snew_RE_RE;
Snew_comp{2,3} = [Snew_RE_RN,Snew_RE_RC(:,4),Snew_RE_LC(:,3),Snew_RE_LN];
Snew_comp{2,4} = Snew_RE_LW;
Snew_comp{2,5} = [Snew_RE_LC(:,1),Snew_RE_RC(:,[2,3]),Snew_RE_LC(:,4)];

Snew_comp{3,1} = [Snew_RN_LS,     Snew_RN_LC(:,2),Snew_RN_RC(:,1),Snew_RN_RS      ;...
                  Snew_RC_LS(4,:),Snew_RC_LC(4,2),Snew_RC_RC(4,1),Snew_RC_RS(4,:) ;...
                  Snew_LC_LS(3,:),Snew_LC_LC(3,2),Snew_LC_RC(3,1),Snew_LC_RS(3,:) ;...
                  Snew_LN_LS,     Snew_LN_LC(:,2),Snew_LN_RC(:,1),Snew_LN_RS     ];
Snew_comp{3,2} = [Snew_RN_RE     ;...
                  Snew_RC_RE(4,:);... 
                  Snew_LC_RE(3,:);... 
                  Snew_LN_RE];
Snew_comp{3,3} = [Snew_RN_RN,     Snew_RN_RC(:,4),Snew_RN_LC(:,3),Snew_RN_LN      ;...
                  Snew_RC_RN(4,:),Snew_RC_RC(4,4),Snew_RC_LC(4,3),Snew_RC_LN(4,:) ;...
                  Snew_LC_RN(3,:),Snew_LC_RC(3,4),Snew_LC_LC(3,3),Snew_LC_LN(3,:) ;...
                  Snew_LN_RN,     Snew_LN_RC(:,4),Snew_LN_LC(:,3),Snew_LN_LN     ];
Snew_comp{3,4} = [Snew_RN_LW     ;...
                  Snew_RC_LW(4,:);... 
                  Snew_LC_LW(3,:);... 
                  Snew_LN_LW];
Snew_comp{3,5} = [Snew_RN_LC(:,1),Snew_RN_RC(:,2:3),Snew_RN_LC(:,4);...
                  Snew_RC_LC(4,1),Snew_RC_RC(4,2:3),Snew_RC_LC(4,4);...
                  Snew_LC_LC(3,1),Snew_LC_RC(3,2:3),Snew_LC_LC(3,4);...
                  Snew_LN_LC(:,1),Snew_LN_RC(:,2:3),Snew_LN_LC(:,4)];
   
Snew_comp{4,1} = [Snew_LW_LS,Snew_LW_LC(:,2),Snew_LW_RC(:,1),Snew_LW_RS];
Snew_comp{4,2} = Snew_LW_RE;
Snew_comp{4,3} = [Snew_LW_RN,Snew_LW_RC(:,4),Snew_LW_LC(:,3),Snew_LW_LN];
Snew_comp{4,4} = Snew_LW_LW;
Snew_comp{4,5} = [Snew_LW_LC(:,1),Snew_LW_RC(:,[2,3]),Snew_LW_LC(:,4)];

Snew_comp{5,1} = [Snew_LC_LS(1,:),Snew_LC_LC(1,2),Snew_LC_RC(1,1),Snew_LC_RS(1,:);...
                  Snew_RC_LS(2,:),Snew_RC_LC(2,2),Snew_RC_RC(2,1),Snew_RC_RS(2,:);...
                  Snew_RC_LS(3,:),Snew_RC_LC(3,2),Snew_RC_RC(3,1),Snew_RC_RS(3,:);...
                  Snew_LC_LS(4,:),Snew_LC_LC(4,2),Snew_LC_RC(4,1),Snew_LC_RS(4,:)];
Snew_comp{5,2} = [Snew_LC_RE(1,:);...
                  Snew_RC_RE(2,:);...
                  Snew_RC_RE(3,:);...
                  Snew_LC_RE(4,:)];
Snew_comp{5,3} = [Snew_LC_RN(1,:),Snew_LC_RC(1,4),Snew_LC_LC(1,3),Snew_LC_LN(1,:);...
                  Snew_RC_RN(2,:),Snew_RC_RC(2,4),Snew_RC_LC(2,3),Snew_RC_LN(2,:);...
                  Snew_RC_RN(3,:),Snew_RC_RC(3,4),Snew_RC_LC(3,3),Snew_RC_LN(3,:);...
                  Snew_LC_RN(4,:),Snew_LC_RC(4,4),Snew_LC_LC(4,3),Snew_LC_LN(4,:)];
Snew_comp{5,4} = [Snew_LC_LW(1,:);...
                  Snew_RC_LW(2,:);...
                  Snew_RC_LW(3,:);...
                  Snew_LC_LW(4,:)];
Snew_comp{5,5} = [Snew_LC_LC(1,1),Snew_LC_RC(1,2),Snew_LC_RC(1,3),Snew_LC_LC(1,4);...
                  Snew_RC_LC(2,1),Snew_RC_RC(2,2),Snew_RC_RC(2,3),Snew_RC_LC(2,4);...
                  Snew_RC_LC(3,1),Snew_RC_RC(3,2),Snew_RC_RC(3,3),Snew_RC_LC(3,4);...
                  Snew_LC_LC(4,1),Snew_LC_RC(4,2),Snew_LC_RC(1,3),Snew_LC_LC(4,4)];

%%% Set the new index vectors.
Snew_comp{6,1} = [SL{6,1},SL{6,5}(2),SR{6,5}(1),SR{6,1}];
Snew_comp{6,2} = SR{6,2};
Snew_comp{6,3} = [SR{6,3},SR{6,5}(4),SL{6,5}(3),SL{6,3}];
Snew_comp{6,4} = SL{6,4};
Snew_comp{6,5} = [SL{6,5}(1),SR{6,5}(2:3),SL{6,5}(4)];
          
return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function Snew_comp = LOCAL_merge_two_vertical(SD,SU,A)

% Extract certain index vectors.
% (These are needed to retrieve/construct needed elements of A.)
JDn = SD{6,3};
JDc = SD{6,5};
JUs = SU{6,1};
JUc = SU{6,5};

% Construct the relevant components of the CENTER matrix.
X_aa     = SD{3,3};
X_bb     = SU{1,1};
X_aa_inv = inv(X_aa);
X_bb_inv = inv(X_bb);
Y_aa     = inv(X_aa - full(A(JDn,JUs)*(X_bb_inv*A(JUs,JDn))));
Y_bb     = inv(X_bb - full(A(JUs,JDn)*(X_aa_inv*A(JDn,JUs))));
Y_ab_rel = -full(X_aa_inv*A(JDn,JUs));
Y_ba_rel = -full(X_bb_inv*A(JUs,JDn));

%%% Compute influence from DOWN-SOUTH
F_a        = Y_aa*SD{3,1};
F_b        = Y_ba_rel*F_a;
Snew_DS_DS = SD{1,1} - SD{1,3}*F_a;
Snew_DE_DS = SD{2,1} - SD{2,3}*F_a;
Snew_DW_DS = SD{4,1} - SD{4,3}*F_a;
Snew_DC_DS = SD{5,1} - SD{5,3}*F_a;
Snew_UE_DS =         - SU{2,1}*F_b;
Snew_UN_DS =         - SU{3,1}*F_b;
Snew_UW_DS =         - SU{4,1}*F_b;
Snew_UC_DS =         - SU{5,1}*F_b;

%%% Compute influence from DOWN-EAST
F_a        = Y_aa*SD{3,2};
F_b        = Y_ba_rel*F_a;
Snew_DS_DE = SD{1,2} - SD{1,3}*F_a;
Snew_DE_DE = SD{2,2} - SD{2,3}*F_a;
Snew_DW_DE = SD{4,2} - SD{4,3}*F_a;
Snew_DC_DE = SD{5,2} - SD{5,3}*F_a;
Snew_UE_DE =         - SU{2,1}*F_b;
Snew_UN_DE =         - SU{3,1}*F_b;
Snew_UW_DE =         - SU{4,1}*F_b;
Snew_UC_DE =         - SU{5,1}*F_b;

%%% Compute influence from DOWN-WEST
F_a        = Y_aa*SD{3,4};
F_b        = Y_ba_rel*F_a;
Snew_DS_DW = SD{1,4} - SD{1,3}*F_a;
Snew_DE_DW = SD{2,4} - SD{2,3}*F_a;
Snew_DW_DW = SD{4,4} - SD{4,3}*F_a;
Snew_DC_DW = SD{5,4} - SD{5,3}*F_a;
Snew_UE_DW =         - SU{2,1}*F_b;
Snew_UN_DW =         - SU{3,1}*F_b;
Snew_UW_DW =         - SU{4,1}*F_b;
Snew_UC_DW =         - SU{5,1}*F_b;

%%% Compute influence from DOWN-CENTER
F_a        = Y_aa*SD{3,5};
F_b        = Y_ba_rel*F_a;
Snew_DS_DC = SD{1,5}    - SD{1,3}*F_a;
Snew_DE_DC = SD{2,5}    - SD{2,3}*F_a;
Snew_DW_DC = SD{4,5}    - SD{4,3}*F_a;
Snew_DC_DC = SD{5,5}    - SD{5,3}*F_a;
Snew_UE_DC =            - SU{2,1}*F_b;
Snew_UN_DC =            - SU{3,1}*F_b;
Snew_UW_DC =            - SU{4,1}*F_b;
Snew_UC_DC = A(JUc,JDc) - SU{5,1}*F_b;

%%% Compute influence from UP-EAST
F_b        = Y_bb*SU{1,2};
F_a        = Y_ab_rel*F_b;
Snew_DS_UE =         - SD{1,3}*F_a;
Snew_DE_UE =         - SD{2,3}*F_a;
Snew_DW_UE =         - SD{4,3}*F_a;
Snew_DC_UE =         - SD{5,3}*F_a;
Snew_UE_UE = SU{2,2} - SU{2,1}*F_b;
Snew_UN_UE = SU{3,2} - SU{3,1}*F_b;
Snew_UW_UE = SU{4,2} - SU{4,1}*F_b;
Snew_UC_UE = SU{5,2} - SU{5,1}*F_b;

%%% Compute influence from UP-NORTH
F_b        = Y_bb*SU{1,3};
F_a        = Y_ab_rel*F_b;
Snew_DS_UN =         - SD{1,3}*F_a;
Snew_DE_UN =         - SD{2,3}*F_a;
Snew_DW_UN =         - SD{4,3}*F_a;
Snew_DC_UN =         - SD{5,3}*F_a;
Snew_UE_UN = SU{2,3} - SU{2,1}*F_b;
Snew_UN_UN = SU{3,3} - SU{3,1}*F_b;
Snew_UW_UN = SU{4,3} - SU{4,1}*F_b;
Snew_UC_UN = SU{5,3} - SU{5,1}*F_b;

%%% Compute influence from UP-WEST
F_b        = Y_bb*SU{1,4};
F_a        = Y_ab_rel*F_b;
Snew_DS_UW =         - SD{1,3}*F_a;
Snew_DE_UW =         - SD{2,3}*F_a;
Snew_DW_UW =         - SD{4,3}*F_a;
Snew_DC_UW =         - SD{5,3}*F_a;
Snew_UE_UW = SU{2,4} - SU{2,1}*F_b;
Snew_UN_UW = SU{3,4} - SU{3,1}*F_b;
Snew_UW_UW = SU{4,4} - SU{4,1}*F_b;
Snew_UC_UW = SU{5,4} - SU{5,1}*F_b;

%%% Compute influence from UP-CENTER
F_b        = Y_bb*SU{1,5};
F_a        = Y_ab_rel*F_b;
Snew_DS_UC =            - SD{1,3}*F_a;
Snew_DE_UC =            - SD{2,3}*F_a;
Snew_DW_UC =            - SD{4,3}*F_a;
Snew_DC_UC = A(JDc,JUc) - SD{5,3}*F_a;
Snew_UE_UC = SU{2,5}    - SU{2,1}*F_b;
Snew_UN_UC = SU{3,5}    - SU{3,1}*F_b;
Snew_UW_UC = SU{4,5}    - SU{4,1}*F_b;
Snew_UC_UC = SU{5,5}    - SU{5,1}*F_b;

Snew_comp = cell(6,5);

Snew_comp{1,1} = Snew_DS_DS;
Snew_comp{1,2} = [Snew_DS_DE,Snew_DS_DC(:,3),Snew_DS_UC(:,2),Snew_DS_UE];
Snew_comp{1,3} = Snew_DS_UN;
Snew_comp{1,4} = [Snew_DS_UW,Snew_DS_UC(:,1),Snew_DS_DC(:,4),Snew_DS_DW];
Snew_comp{1,5} = [Snew_DS_DC(:,[1,2]),Snew_DS_UC(:,[3,4])];
       
Snew_comp{2,1} = [Snew_DE_DS;...
                  Snew_DC_DS(3,:);...
                  Snew_UC_DS(2,:);...
                  Snew_UE_DS];
Snew_comp{2,2} = [Snew_DE_DE,     Snew_DE_DC(:,3),Snew_DE_UC(:,2),Snew_DE_UE      ;...
                  Snew_DC_DE(3,:),Snew_DC_DC(3,3),Snew_DC_UC(3,2),Snew_DC_UE(3,:) ;...
                  Snew_UC_DE(2,:),Snew_UC_DC(2,3),Snew_UC_UC(2,2),Snew_UC_UE(2,:) ;...
                  Snew_UE_DE     ,Snew_UE_DC(:,3),Snew_UE_UC(:,2),Snew_UE_UE     ];
Snew_comp{2,3} = [Snew_DE_UN;...
                  Snew_DC_UN(3,:);...
                  Snew_UC_UN(2,:);...
                  Snew_UE_UN];
Snew_comp{2,4} = [Snew_DE_UW,     Snew_DE_UC(:,1),Snew_DE_DC(:,4),Snew_DE_DW      ;...
                  Snew_DC_UW(3,:),Snew_DC_UC(3,1),Snew_DC_DC(3,4),Snew_DC_DW(3,:) ;...
                  Snew_UC_UW(2,:),Snew_UC_UC(2,1),Snew_UC_DC(2,4),Snew_UC_DW(2,:) ;...
                  Snew_UE_UW     ,Snew_UE_UC(:,1),Snew_UE_DC(:,4),Snew_UE_DW     ];
Snew_comp{2,5} = [Snew_DE_DC(:,[1,2]),Snew_DE_UC(:,[3,4]) ;...
                  Snew_DC_DC(3,[1,2]),Snew_DC_UC(3,[3,4]) ;...
                  Snew_UC_DC(2,[1,2]),Snew_UC_UC(2,[3,4]) ;...
                  Snew_UE_DC(:,[1,2]),Snew_UE_UC(:,[3,4])];

Snew_comp{3,1} = Snew_UN_DS;
Snew_comp{3,2} = [Snew_UN_DE,Snew_UN_DC(:,3),Snew_UN_UC(:,2),Snew_UN_UE];
Snew_comp{3,3} = Snew_UN_UN;
Snew_comp{3,4} = [Snew_UN_UW,Snew_UN_UC(:,1),Snew_UN_DC(:,4),Snew_UN_DW];
Snew_comp{3,5} = [Snew_UN_DC(:,[1,2]),Snew_UN_UC(:,[3,4])];

Snew_comp{4,1} = [Snew_UW_DS;...
                  Snew_UC_DS(1,:);...
                  Snew_DC_DS(4,:);...
                  Snew_DW_DS];
Snew_comp{4,2} = [Snew_UW_DE,     Snew_UW_DC(:,3),Snew_UW_UC(:,2),Snew_UW_UE      ;...
                  Snew_UC_DE(1,:),Snew_UC_DC(1,3),Snew_UC_UC(1,2),Snew_UC_UE(1,:) ;...
                  Snew_DC_DE(4,:),Snew_DC_DC(4,3),Snew_DC_UC(4,2),Snew_DC_UE(4,:) ;...
                  Snew_DW_DE     ,Snew_DW_DC(:,3),Snew_DW_UC(:,2),Snew_DW_UE     ];
Snew_comp{4,3} = [Snew_UW_UN;...
                  Snew_UC_UN(1,:);...
                  Snew_DC_UN(4,:);...
                  Snew_DW_UN];
Snew_comp{4,4} = [Snew_UW_UW,     Snew_UW_UC(:,1),Snew_UW_DC(:,4),Snew_UW_DW      ;...
                  Snew_UC_UW(1,:),Snew_UC_UC(1,1),Snew_UC_DC(1,4),Snew_UC_DW(1,:) ;...
                  Snew_DC_UW(4,:),Snew_DC_UC(4,1),Snew_DC_DC(4,4),Snew_DC_DW(4,:) ;...
                  Snew_DW_UW     ,Snew_DW_UC(:,1),Snew_DW_DC(:,4),Snew_DW_DW     ];
Snew_comp{4,5} = [Snew_UW_DC(:,[1,2]),Snew_UW_UC(:,[3,4]) ;...
                  Snew_UC_DC(1,[1,2]),Snew_UC_UC(1,[3,4]) ;...
                  Snew_DC_DC(4,[1,2]),Snew_DC_UC(4,[3,4]) ;...
                  Snew_DW_DC(:,[1,2]),Snew_DW_UC(:,[3,4])];

Snew_comp{5,1} = [Snew_DC_DS(1,:);...
                  Snew_DC_DS(2,:);...
                  Snew_UC_DS(3,:);...
                  Snew_UC_DS(4,:)];
Snew_comp{5,2} = [Snew_DC_DE(1,:),Snew_DC_DC(1,3),Snew_DC_UC(1,2),Snew_DC_UE(1,:);...
                  Snew_DC_DE(2,:),Snew_DC_DC(2,3),Snew_DC_UC(2,2),Snew_DC_UE(2,:);...
                  Snew_UC_DE(3,:),Snew_UC_DC(3,3),Snew_UC_UC(3,2),Snew_UC_UE(3,:);...
                  Snew_UC_DE(4,:),Snew_UC_DC(4,3),Snew_UC_UC(4,2),Snew_UC_UE(4,:)];
Snew_comp{5,3} = [Snew_DC_UN(1,:);...
                  Snew_DC_UN(2,:);...
                  Snew_UC_UN(3,:);...
                  Snew_UC_UN(4,:)];
Snew_comp{5,4} = [Snew_DC_UW(1,:),Snew_DC_UC(1,1),Snew_DC_DC(1,4),Snew_DC_DW(1,:);...
                  Snew_DC_UW(2,:),Snew_DC_UC(2,1),Snew_DC_DC(2,4),Snew_DC_DW(2,:);...
                  Snew_UC_UW(3,:),Snew_UC_UC(3,1),Snew_UC_DC(3,4),Snew_UC_DW(3,:);...
                  Snew_UC_UW(4,:),Snew_UC_UC(4,1),Snew_UC_DC(4,4),Snew_UC_DW(4,:)];
Snew_comp{5,5} = [Snew_DC_DC(1,1),Snew_DC_DC(1,2),Snew_DC_UC(1,3),Snew_DC_UC(1,4);...
                  Snew_DC_DC(2,1),Snew_DC_DC(2,2),Snew_DC_UC(2,3),Snew_DC_UC(2,4);...
                  Snew_UC_DC(3,1),Snew_UC_DC(3,2),Snew_UC_UC(3,3),Snew_UC_UC(3,4);...
                  Snew_UC_DC(4,1),Snew_UC_DC(4,2),Snew_UC_UC(4,3),Snew_UC_UC(4,4)];

%%% Set the new index vectors.
Snew_comp{6,1} = SD{6,1};
Snew_comp{6,2} = [SD{6,2},SD{6,5}(3),SU{6,5}(2),SU{6,2}];
Snew_comp{6,3} = SU{6,3};
Snew_comp{6,4} = [SU{6,4},SU{6,5}(1),SD{6,5}(4),SD{6,4}];
Snew_comp{6,5} = [SD{6,5}(1:2),SU{6,5}(3:4)];
      
return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [Jsw,Js,Jse,Je,Jne,Jn,Jnw,Jw,Jrim] = LOCAL_get_J_pieces(NODES,ibox)

ilevel = NODES{ 2,ibox};
ncell  = NODES{ 9,ibox};
i1     = NODES{10,ibox};
i2     = NODES{11,ibox};

ncol   = ncell * (2^ilevel);
ifirst = ncol*ncell*(i1-1) + ncell*(i2-1) + 1;

Jrim  = [ifirst + ncol*(0:(ncell-1)),...
         ifirst + ncol*(ncell-1) + (1:(ncell-1)),...
         ncol*(i1*ncell - 1) + ncell*i2 + ncol*((-1):(-1):(-ncell+1)),...
         ifirst + ((ncell-2):(-1):1)];

Jsw = ifirst;
Js  = ifirst + ncol*(1:(ncell-2));
Jse = ifirst + ncol*(ncell-1);
Je  = ifirst + ncol*(ncell-1) + (1:(ncell-2));
Jne = ifirst + ncol*(ncell-1) + ncell - 1;
Jn  = ncol*(i1*ncell - 1) + ncell*i2 + ncol*((-1):(-1):(-ncell+2));
Jnw = ifirst + ncell - 1;
Jw  = ifirst + ((ncell-2):(-1):1);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function J_int = LOCAL_get_Jint(NODES,ibox)

ilevel = NODES{ 2,ibox};
ncell  = NODES{ 9,ibox};
i1     = NODES{10,ibox};
i2     = NODES{11,ibox};

ncol   = ncell * (2^ilevel);
J_int  = zeros(1,(ncell-2)*(ncell-2));

for i = 1:(ncell-2)
  J_int((i-1)*(ncell-2) + (1:(ncell-2))) = ...
      ncol*(ncell*(i1-1)+i) + ncell*(i2-1) + (2:(ncell-1));
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function B = LOCAL_get_S_from_S_comp(Bxxx)

Bss = Bxxx{1,1};
Bse = Bxxx{1,2};
Bsn = Bxxx{1,3};
Bsw = Bxxx{1,4};
Bsc = Bxxx{1,5};

Bes = Bxxx{2,1};
Bee = Bxxx{2,2};
Ben = Bxxx{2,3};
Bew = Bxxx{2,4};
Bec = Bxxx{2,5};

Bns = Bxxx{3,1};
Bne = Bxxx{3,2};
Bnn = Bxxx{3,3};
Bnw = Bxxx{3,4};
Bnc = Bxxx{3,5};

Bws = Bxxx{4,1};
Bwe = Bxxx{4,2};
Bwn = Bxxx{4,3};
Bww = Bxxx{4,4};
Bwc = Bxxx{4,5};

Bcs = Bxxx{5,1};
Bce = Bxxx{5,2};
Bcn = Bxxx{5,3};
Bcw = Bxxx{5,4};
Bcc = Bxxx{5,5};

B = [Bcc(1,1), Bcs(1,:), Bcc(1,2), Bce(1,:), Bcc(1,3), Bcn(1,:), Bcc(1,4), Bcw(1,:);...
     Bsc(:,1), Bss,      Bsc(:,2), Bse,      Bsc(:,3), Bsn,      Bsc(:,4), Bsw;...
     Bcc(2,1), Bcs(2,:), Bcc(2,2), Bce(2,:), Bcc(2,3), Bcn(2,:), Bcc(2,4), Bcw(2,:);...
     Bec(:,1), Bes,      Bec(:,2), Bee,      Bec(:,3), Ben,      Bec(:,4), Bew;...
     Bcc(3,1), Bcs(3,:), Bcc(3,2), Bce(3,:), Bcc(3,3), Bcn(3,:), Bcc(3,4), Bcw(3,:);...
     Bnc(:,1), Bns,      Bnc(:,2), Bne,      Bnc(:,3), Bnn,      Bnc(:,4), Bnw;...
     Bcc(4,1), Bcs(4,:), Bcc(4,2), Bce(4,:), Bcc(4,3), Bcn(4,:), Bcc(4,4), Bcw(4,:);...
     Bwc(:,1), Bws,      Bwc(:,2), Bwe,      Bwc(:,3), Bwn,      Bwc(:,4), Bww];
 
 return

 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 function [T, I] = LOCAL_id_decomp(A,INPUT2)

if (INPUT2 < 1)
  acc = INPUT2;
  [T,I] = LOCAL_skel_col2(A,acc);
else
  k = INPUT2;
  [T,I] = LOCAL_skel_col3(A,k);
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [T, I] = LOCAL_skel_col2(A,acc)

ss = svd(A);
k  = sum(ss > acc);
[tmp, R, I] = qr(A,0);

[U,D,V] = svd(R(1:k,1:k));
q = sum(diag(D) > acc);
T = V(:,1:q)*(D(1:q,1:q)\(U(:,1:q)'))*R(1:k, (k+1):size(R,2));

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [T, I] = LOCAL_skel_col3(A,k)

[tmp, R, I] = qr(A,0);
[U,D,V] = svd(R(1:k,1:k));
q = sum(diag(D) > 1e-12);
T = V(:,1:q)*(D(1:q,1:q)\(U(:,1:q)'))*R(1:k, (k+1):size(R,2));

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

